#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <cassert>
#include <cmath>

#include "vector.h"
#include "pixel.h"
#include "camera.h"
#include "light.h"
#include "material.h"
#include "object.h"
#include "list.h"
#include "plane.h"
#include "model.h"
#include "ray.h"

void ray_t::trace(model_t&		model,
                  rgb_t<double>&	color,
                  object_t		*last_hit)
{
	object_t			*obj=NULL;
	rgb_t<double>			ambient, diffuse, specular;
	material_t			*mat;
	double				rdis=0.0;
	vec_t				N;

	light_t				*lgt=NULL;
	vec_t				L,V,R,H;
	rgb_t<double>			I_d, I_s;
	double				r,ndotl=0.0,n=32.0;
	list_t<light_t* >::iterator	litr;

  // get closest object, if any
  if(!(obj = model.find_closest(pos,dir,last_hit,rdis)) || dis > MAX_DIST)
    return;

  // if hit, print info
  if(rdis > 0) {

    // accumulate distance travelled by ray
    dis += rdis;

    // get object material properties
    if((mat = model.getmaterial(obj->getmaterial())) != NULL) {
      ambient = mat->getamb();
      diffuse = mat->getdiff();
      specular = mat->getspec();
    }

    // ambient color
    color += 1.0/dis * ambient;	// ambient scaled by ray dist

    // clamp resultant color
    color.clamp(0.0,1.0);

    N = obj->getnormal().norm();// surf normal

    // view direction (direction from hit point to camera)
    V = -dir;			// dir to camera

    // diffuse component from each light...
    if(!diffuse.iszero()) {
      for(litr = model.lgts.begin(); litr != model.lgts.end(); litr++) {
        // pointer to light
        lgt = (light_t *)*litr;
  
        // light direction and distance
        L = lgt->getlocation() - obj->getlast_hit();
        r = L.len();				// distance to light
        L = L.norm();				// dir to light
  
        // angle with light
        ndotl = N.dot(L);
  
        // check visibility wrt light
        if(0.0 < ndotl && ndotl < 1.0) {
          // light color scaled by N . L
          I_d = 1.0/r * ndotl * lgt->getcolor();
  
          // add in diffuse contribution from light scaled by material property
          color += (I_d * diffuse);
        }
      }
  
      // clamp resultant color
      color.clamp(0.0,1.0);
    }

    // reflection
    if(!specular.iszero()) {
	rgb_t<double>	refcolor;
	vec_t		r = dir.reflect(N);
	ray_t		*reflection = new ray_t(obj->getlast_hit(),r,dis);

      // trace the reflection
      reflection->trace(model,refcolor,obj);
      delete reflection;

      // composite surface color: blending baesd on surface properties
      // (note that diffuse + specular should add up to 1)
      color  = (diffuse * color) + (specular * refcolor);

      // clamp resultant color
      color.clamp(0.0,1.0);
    }

    // specular highlights
    if(!specular.iszero()) {

      // add in specular highlight from each light...
      for(litr = model.lgts.begin(); litr != model.lgts.end(); litr++) {
        // pointer to light
        lgt = (light_t *)*litr;

        // light direction and distance
        L = lgt->getlocation() - obj->getlast_hit();
        r = L.len();				// distance to light
        L = L.norm();				// dir to light

        // angle with light
        ndotl = N.dot(L);

        // check visibility wrt light
        if(0.0 < ndotl && ndotl < 1.0) {
          // specular reflection direction (and bisector)
          R = L.reflect(N);
          // bisector (not used)
          H = (0.5 * (L + V)).norm();

          // light color scaled by (R . V)^n
          I_s = 1.0/r * pow(R.dot(V),n) * lgt->getcolor();
  
          // add in specular contribution from light scaled by material property
          color += (I_s * specular);
        }

        // clamp resultant color
        color.clamp(0.0,1.0);
      }
    }
  }
}
